/*
** 1998-05-25 -	The amount of code in the main module dealing with dirpanes simply got too big. Dike, dike.
** 1998-06-08 -	Redesigned sorting somewhat. Now, the information used to determine how to sort resides
**		in the configuration structure, not the individual dir pane.
** 1998-08-02 -	Fixed big performance mishap; when changing the sort mode, a full dir reread was done,
**		rather than just a (relatively quick) redisplay.
** 1998-08-08 -	Now finally supports high-speed dragging without losing lines. Great.
** 1999-03-05 -	Massive changes since we now rely on GTK+'s CList widget to handle all selection details.
**		Gives dragging, scrolling, and stuff.
** 1999-03-14 -	Opaque-ified the access to DirRow fields.
** 1999-05-29 -	Added support for symlinks. More controlled and more memory efficient than the old code.
** 2000-04-16 -	Simplified sorting code somewhat, implemented actual comparison functions for {u,g}name.
*/

#include "gentoo.h"

#include <ctype.h>
#include <glob.h>
#include <stdlib.h>
#include <string.h>

#include <gdk/gdkkeysyms.h>

#include "children.h"
#include "cmdseq.h"
#include "configure.h"
#include "controls.h"
#include "convutil.h"
#include "dirhistory.h"
#include "dpformat.h"
#include "errors.h"
#include "events.h"
#include "fileutil.h"
#include "gfam.h"
#include "guiutil.h"
#include "mount.h"
#include "sizeutil.h"
#include "strutil.h"
#include "styles.h"
#include "types.h"
#include "userinfo.h"

#include "cmd_dpfocus.h"

#include "dirpane.h"

/* ----------------------------------------------------------------------------------------- */

typedef gint	(*StrCmpFunc)(const gchar *a, const gchar *b);

static StrCmpFunc	the_strcmp_case = NULL, the_strcmp_nocase = NULL;

static DPSort	the_sort;			/* Used to get state across to qsort() callback. */
static gint	(*the_strcmp)(const gchar *a, const gchar *b);

static GtkStyle	*pane_selected = NULL,		/* The style used on selected pane's column buttons. */
		*focus_style   = NULL;		/* For focusing. */

static gboolean	no_repeat = FALSE;		/* Ugly hack to prevent repeat on focused row to go into infinity. Ugly. Ugly. */

/* This is used by dp_activate_queued() to keep track of state. */
static struct {
	guint	handler;
	DirPane	*old_cur;
} the_activate_queue_info = { 0, NULL };

static GSList	*pathwidgetry_builders = NULL;

/* ----------------------------------------------------------------------------------------- */

#include "graphics/icon_dirparent.xpm"		/* Graphics for the parent button. */

/* ----------------------------------------------------------------------------------------- */

static void	clear_total_stats(DirPane *dp);
static void	clear_selection_stats(SelInfo *si);
static void	update_selection_stats(DirPane *dp);

static gint	numeric_strcmp(const gchar *a, const gchar *b);
static gint	numeric_strcasecmp(const gchar *a, const gchar *b);

static void	evt_dirpane_select_row(GtkWidget *wid, gint row, gint col, GdkEventButton *ev, gpointer user);
static void	evt_dirpane_unselect_row(GtkWidget *wid, gint row, gint col, GdkEventButton *ev, gpointer user);

static void	dp_activate_queued(DirPane *dp);

/* ----------------------------------------------------------------------------------------- */

#if defined ENABLE_NLS
static gint cmp_wrapper(gconstpointer a, gconstpointer b)
{
	return strcoll(*(const gchar **) a, *(const gchar **) b);
}
#endif

/* 2002-06-17 -	Find out which string comparison function to use, and when. This information
**		is then used by the logic that drives the pane sorting. Complicated, since
**		people seem to implement strcoll() in wildly different ways. *Sigh*.
*/
void dp_initialize(DirPane *dp)
{
#if defined ENABLE_NLS
	const gchar	*locale, *test[] = { "c", "a", "B" };
#endif

	/* Clear the path settings. */
	dp[0].dir.path[0] = '\0';
	dp[0].dir.pathd = NULL;
	dp[1].dir.path[0] = '\0';
	dp[1].dir.pathd = NULL;

	/* Assume no localization support at all. Ooh, the easy life. */
	the_strcmp_case   = strcmp;
	the_strcmp_nocase = g_strcasecmp;

#if defined ENABLE_NLS
	/* If we're internationalized, we need to determine if strcoll() behaves or not. */

	/* Query the collation setting, blatantly assuming that's what controls strcoll(). */
	if((locale = setlocale(LC_COLLATE, NULL)) == NULL)
		return;

	/* Let's be a bit brute-forceish, here. Sort a sample array, then inspect. */
	qsort(test, sizeof test / sizeof *test, sizeof *test, cmp_wrapper);

	/* Is the order "a", "B", "c"? Then that means strcoll() ignores case, which
	** in turn means that case-sensitive comparisons will not work correctly.
	*/
	if(test[0][0] == 'a' && test[1][0] == 'B')
	{
		/* This means strcoll() isn't case-sensitive. Inform user. */
		the_strcmp_case   = strcmp;		/* Won't work as expected. */
		the_strcmp_nocase = strcoll;
		g_warning(_("Case-sensitive sorting will not work correctly with non-ASCII characters"));
	}
	else if(strcmp(locale, "C") != 0 && strcmp(locale, "POSIX") != 0)
	{
		/* If we're here, strcoll() cares about case. Which is weird, too,
		** since we can't invent a case-insensitive variant on our own.
		*/
		the_strcmp_case   = strcoll;
		the_strcmp_nocase = g_strcasecmp;	/* Won't work as expected. */
		g_warning(_("Case-insensitive sorting will not work correctly with non-ASCII characters"));
	}
#endif
}

/* ----------------------------------------------------------------------------------------- */

/* 2002-08-02 -	Rewrote this classic, to be a bit shorter and take const-ant time, too. Whoo. */
DirPane * dp_mirror(const MainInfo *min, const DirPane *dp)
{
	return &min->gui->pane[1 - dp->index];
}

/* ----------------------------------------------------------------------------------------- */

/* 1998-09-05 -	Get the full name, with path, for row number <row> of <dp>. Trivial, but
**		saves a couple of lines here and there elsewhere in the program.
** 1999-01-20 -	Now tries to avoid starting the name with two slashes for root entries.
** 2002-02-25 -	Generalized to never produce double slashes, removed special case for root.
*/
const gchar * dp_full_name(const DirPane *dp, gint row)
{
	static gchar	buf[PATH_MAX];
	const gchar	*sep = "";
	gint		len;

	if((len = strlen(dp->dir.path)) > 0 && dp->dir.path[len - 1] != G_DIR_SEPARATOR)
		sep = G_DIR_SEPARATOR_S;
	g_snprintf(buf, sizeof buf, "%s%s%s", dp->dir.path, sep, DP_ROW_NAME(&dp->dir.row[row]));
	return buf;
}

/* 1999-02-23 -	Get the name of a row, but quoted and with any embedded quotes
**		escaped by backslashes. Very handy when evaluating {f}-codes...
**		If <path> is TRUE, the path is included as well. Doesn't nest.
*/
const gchar * dp_name_quoted(const DirPane *dp, gint row, gboolean path)
{
	static gchar	buf[2 * PATH_MAX];
	const gchar	*fn, *ptr;
	gchar		*put = buf;

	fn = path ? dp_full_name(dp, row) : DP_ROW_NAME(&dp->dir.row[row]);
	*put++ = '"';
	for(ptr = fn; *ptr; ptr++)
	{
		if(*ptr == '"' || *ptr == '\\')
			*put++ = '\\';
		*put++ = *ptr;
	}
	*put++ = '"';
	*put = '\0';

	return buf;
}

/* ----------------------------------------------------------------------------------------- */

/* 2008-09-20 -	Clear the completion cache. */
static void completion_clear(DirPane *dp)
{
	GtkListStore	*cm;

	cm = GTK_LIST_STORE(gtk_entry_completion_get_model(dp->complete.compl));
	if(cm == NULL)
		return;
	gtk_list_store_clear(cm);
}

/* 2008-09-20 -	Update the completion for the given pane, considering that 'prefix' is the
 *		current directory root. If this is the same as the cached, do nothing.
*/
static void completion_update(DirPane *dp, const gchar *prefix)
{
	GtkListStore	*cm;
	GDir		*dir;
	Conv		conv;
	gchar		*fs;

	if(strcmp(dp->complete.prefix, prefix) == 0)
		return;
	cm = GTK_LIST_STORE(gtk_entry_completion_get_model(dp->complete.compl));
	if(cm == NULL)
		return;
	/* We're about to check on-disk, so clear the cache in the GtkEntryCompletion. */
	gtk_list_store_clear(cm);
	/* Convert UTF-8 prefix string to filesystem encoding, and open directory. */
	conv_init(&conv);
	fs = conv_begin_reverse(&conv, prefix);
	if((dir = g_dir_open(fs, 0u, NULL)) != NULL)
	{
		const gchar	*name;

		while((name = g_dir_read_name(dir)) != NULL)
		{
			gchar		fname[4 * PATH_MAX];
			GtkTreeIter	iter;

			/* Note: pre INCLUDES trailing slash. Always. */
			g_snprintf(fname, sizeof fname, "%s%s", prefix, name);
			if(g_file_test(fname, G_FILE_TEST_IS_DIR))
			{
				/* Got a directory, now upconvert its name to UTF-8 and store. */
				Conv	conv;
				gchar	*dpl;

				dpl = conv_begin_reverse(&conv, fname);
				gtk_list_store_append(cm, &iter);
				gtk_list_store_set(cm, &iter, 0, dpl, -1);
				conv_end(&conv);
			}
		}
		g_strlcpy(dp->complete.prefix, prefix, sizeof dp->complete.prefix);
		g_dir_close(dir);
	}
	conv_end(&conv);
}

static gboolean completion_match_function(GtkEntryCompletion *completion, const gchar *key, GtkTreeIter *iter, gpointer user)
{
	gchar	*mv;

	gtk_tree_model_get(gtk_entry_completion_get_model(completion), iter, 0, &mv, -1);
	if(mv != NULL)
	{
		gsize		klen = strlen(key);	/* NOTE: This is in *bytes*, not UTF-8 chars. */
		gboolean	eq = memcmp(key, mv, klen) == 0;

		g_free(mv);
		return eq;
	}
	return FALSE;
}

/* ----------------------------------------------------------------------------------------- */

/* 2000-02-03 -	Here's the handler for the optional "huge parent" button, in the pane margin. */
static void evt_hugeparent_clicked(GtkWidget *wid, gpointer user)
{
	DirPane	*dp = user;

	csq_execute(dp->main, dp->index == 0 ? "ActivateLeft" : "ActivateRight");
	csq_execute(dp->main, "DirParent");
}

/* 1998-09-15 -	This handler gets run when user clicks the "up" button next to path entry. */
static void evt_parent_clicked(GtkWidget *wid, gpointer user)
{
	DirPane	*dp = user;

	csq_execute(dp->main, dp->index == 0 ? "ActivateLeft" : "ActivateRight");
	csq_execute(dp->main, "DirParent");
}

/* 2008-09-13 -	A simple rewrite of the older code, considering the switch to a GtkComboBoxEntry
 *		widget for the path entry/history tasks.
*/
static void evt_path_new(GtkWidget *wid, gpointer user)
{
	DirPane	*dp = user;
	Conv	conv;
	gchar	*dir;

	dp_activate(dp);
	/* Convert user's text into filesystem encoding. */
	conv_init(&conv);
	dir = conv_begin(&conv, gtk_combo_box_get_active_text(GTK_COMBO_BOX(dp->path)));
	csq_execute_format(dp->main, "DirEnter 'dir=%s'", stu_escape(dir));
	conv_end(&conv);
	dp_path_unfocus(dp);
}

static void evt_path_changed(GtkWidget *wid, gpointer user)
{
	gint	index;
	gchar	*now;

	now = gtk_combo_box_get_active_text(GTK_COMBO_BOX(wid));
	index = gtk_combo_box_get_active(GTK_COMBO_BOX(wid));
	if(index < 0)
	{
		gchar	*ptr;

		/* Build "prefix" version of the current path. The prefix is defined, for completion
		 * purposes, as the text up to and including the right-most directory separator. We
		 * do this in non-dynamic storage, for speed.
		*/
		ptr = g_utf8_strrchr(now, -1, G_DIR_SEPARATOR);
		if(ptr != NULL)
		{
			gchar	buf[1024];
			size_t	plen = ptr - now + 1;
			if(plen < sizeof buf)
			{
				Conv	conv;
				gchar	*disk;

				memcpy(buf, now, plen);
				buf[plen] = '\0';
				/* Now, we have a UTF-8 prefix in buf. Convert to filesystem encoding. */
				conv_init(&conv);
				if((disk = conv_begin(&conv, buf)) != NULL)
					completion_update(user, buf);
				conv_end(&conv);
			}
		}
	}
	else
		evt_path_new(wid, user);
	g_free(now);
}

/* 2008-09-20 -	Cut away Tab handling, we now use the slower but more modern complete-as-you type style. */
static gboolean evt_path_key_press(GtkWidget *wid, GdkEventKey *evt, gpointer user)
{
	DirPane	*dp = user;

	if(evt->keyval == GDK_Escape)
		dp_path_unfocus(dp);
	return FALSE;
}

/* 2004-04-18 -	Mouse click in path entry widget. */
static gboolean evt_path_button_press(GtkWidget *wid, GdkEventButton *evt, gpointer user)
{
	if(evt->type == GDK_BUTTON_PRESS && evt->button == 3)
	{
		DirPane		*dp = user;
		const gchar	*cmd;

		/* If a non-empty command binding is found, run it and consume event. */
		if((cmd = ctrl_general_get_cmdseq(dp->main->cfg.ctrlinfo, CTRL_GENERAL_PATH_RMB)) != NULL && *cmd != '\0')
		{
			csq_execute(dp->main, cmd);
			return TRUE;
		}
	}
	return FALSE;
}

/* 1998-05-27 -	I had forgotten about the possibility for users to click in the path entry widget, and
**		by doing so circumventing my clever (hairy) flag system. This remedies that.
** 1998-05-29 -	Changed polarity of operation. Was broken. Sloppy me.
*/
static gboolean evt_path_focus(GtkWidget *wid, GdkEventFocus *ev, gpointer user)
{
	DirPane	*dp = user;

	kbd_context_detach(dp->main->gui->kbd_ctx, GTK_WINDOW(dp->main->gui->window));
	dp_activate(dp);

	g_signal_connect(G_OBJECT(wid), "key_press_event", G_CALLBACK(evt_path_key_press), dp);

	return FALSE;
}

static gboolean evt_path_unfocus(GtkWidget *wid, GdkEventFocus *ev, gpointer user)
{
	DirPane	*dp = user;

	kbd_context_attach(dp->main->gui->kbd_ctx, GTK_WINDOW(dp->main->gui->window));

	gtk_signal_disconnect_by_func(GTK_OBJECT(wid), GTK_SIGNAL_FUNC(evt_path_key_press), dp);

	return FALSE;
}

/* 1998-12-19 -	The tiny little "hide" button was clicked. Act. */
static void evt_hide_clicked(GtkWidget *wid, gpointer user)
{
	DirPane	*dp = user;

	dp_activate(dp);
	dp->main->cfg.dp_format[dp->index].hide_allowed = GTK_TOGGLE_BUTTON(wid)->active;
	csq_execute(dp->main, "DirRescan");
}

/* ----------------------------------------------------------------------------------------- */

/* 1998-09-10 -	Nestable freeze. Very useful; ALWAYS use these rather than direct
**		freezing/thawing on the dp's clist!
** 1999-03-04 -	GTK+ 1.2.0 does the right thing, making this function even simpler.
*/
void dp_freeze(DirPane *dp)
{
	gtk_clist_freeze(GTK_CLIST(dp->list));
}

/* 1998-09-10 -	Nestable thaw. Always use it!
** 1998-09-19 -	Umm... This function was just proved to be affecting (clearing) errno.
**		That fscks up the error handling all over the place, so I added some
**		local save/restore of it. I can't claim to understand any of it, though.
** 1999-03-04 -	Did GTK+ 1.2.0 simplifications.
*/
void dp_thaw(DirPane *dp)
{
	gint	old_errno = errno;

	gtk_clist_thaw(GTK_CLIST(dp->list));
	errno = old_errno;
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-03-06 -	The given <dp> has been double-clicked. Time to execute some fun action,
**		namely the Default for the style of the clicked row.
*/
static void doubleclick(DirPane *dp)
{
/*	err_clear(dp->main);*/
	csq_execute(dp->main, "FileAction");
	if(chd_get_running(NULL) == NULL)	/* If command sequence finished, reset. Else keep around. */
		dp->dbclk_row = -1;
/*	err_show(dp->main);*/
}

/* 1999-05-13 -	Simulate a double click on <row> in <dp>. Handy for use by the focusing
**		module.
*/
void dp_dbclk_row(DirPane *dp, gint row)
{
	if(row != -1)
	{
		dp->dbclk_row = row;
		doubleclick(dp);
	}
}

/* 1999-03-04 -	Rewritten. Now very much simpler. :) */
void dp_select(DirPane *dp, gint row)
{
	if(g_list_find(GTK_CLIST(dp->list)->selection, GINT_TO_POINTER(row)) != NULL)
		return;
	gtk_clist_select_row(GTK_CLIST(dp->list), row, -1);
}

/* 1999-03-04 -	Select all rows. Handy for the SelectAll command. Since selection management
**		is now largely done by the GtkCList widget for us, this is not rocket science.
** 2000-09-16 -	Noticed that using gtk_clist_select_all() reset the vertical scroll of the list
**		to zero, which annoyed me. This happens regardless of callback. Did a work around.
*/
void dp_select_all(DirPane *dp)
{
	guint	i;

	dp_freeze(dp);
	gtk_signal_handler_block_by_func(GTK_OBJECT(dp->list), GTK_SIGNAL_FUNC(evt_dirpane_unselect_row), dp);
	/* Can't use gtk_clist_select_all(); it messes up the vertical position. Weird. */
        for(i = 0; i < dp->dir.num_rows; i++)
                gtk_clist_select_row(GTK_CLIST(dp->list), i, -1);
	gtk_signal_handler_unblock_by_func(GTK_OBJECT(dp->list), GTK_SIGNAL_FUNC(evt_dirpane_unselect_row), dp);
	dp_thaw(dp);
	update_selection_stats(dp);
	dp_show_stats(dp);
}

/* 1999-03-04 -	Rewritten in a lot simpler way. */
void dp_unselect(DirPane *dp, gint row)
{
	if(row == dp->dbclk_row)
		dp->dbclk_row = -1;
	else
		gtk_clist_unselect_row(GTK_CLIST(dp->list), row, 0);
}

/* 1999-03-04 -	Unselect all rows. Real simple. */
void dp_unselect_all(DirPane *dp)
{
	dp->dbclk_row = -1;
	dp_freeze(dp);
	gtk_signal_handler_block_by_func(GTK_OBJECT(dp->list), GTK_SIGNAL_FUNC(evt_dirpane_unselect_row), dp);
	gtk_clist_unselect_all(GTK_CLIST(dp->list));
	gtk_signal_handler_unblock_by_func(GTK_OBJECT(dp->list), GTK_SIGNAL_FUNC(evt_dirpane_unselect_row), dp);
	dp_thaw(dp);
	update_selection_stats(dp);
	dp_show_stats(dp);
}

/* 2000-03-18 -	Since the select/unselect functions no longer return success or failure, the check
**		is done explicitly here instead. Much better, really.
*/
void dp_toggle(DirPane *dp, gint row)
{
	if(g_list_find(GTK_CLIST(dp->list)->selection, GINT_TO_POINTER(row)) == NULL)
		dp_select(dp, row);
	else
		dp_unselect(dp, row);
}

static void focus_select(DirPane *dp, gint row)
{
	if((dp->focus_row != -1) && dpf_get_focus_select())
		dp_focus(dp, row);
}

/* 1999-03-04 -	A row in the pane was selected. Update statistics. */
static void evt_dirpane_select_row(GtkWidget *wid, gint row, gint col, GdkEventButton *ev, gpointer user)
{
	DirPane	*dp = user;
	DirRow	*dr;

	if((dr = gtk_clist_get_row_data(GTK_CLIST(wid), row)) == NULL)
		return;

	if(S_ISDIR(DP_ROW_LSTAT(dr).st_mode))
		dp->dir.sel.num_dirs++;
	else
		dp->dir.sel.num_files++;

	dp->dir.sel.num_bytes  += DP_ROW_LSTAT(dr).st_size;
	dp->dir.sel.num_blocks += DP_ROW_LSTAT(dr).st_blocks;
	dp_show_stats(dp);

	focus_select(dp, row);

	if(dp->dbclk_row == row)
		doubleclick(dp);
}

/* 1999-03-04 -	A row has been unselected, so we need to update our statistics.
** 1999-03-06 -	Now does the double-click "activation". When we arrive here, the double-
**		clicked row is no longer selected (dp_get_selection() will not include it),
**		which is very important, for example for the directory history.
*/
static void evt_dirpane_unselect_row(GtkWidget *wid, gint row, gint col, GdkEventButton *ev, gpointer user)
{
	DirPane	*dp = user;

	update_selection_stats(dp);
	dp_show_stats(dp);

	if(dp->dbclk_row >= 0)		/* Did a double-click occur resently? */
		doubleclick(dp);
}

/* 1999-03-05 -	Update the state of the GDK_CONTROL_MASK bitmask in <state>, depending on how
**		the given pane has its ctrl_mode formatting mode set.
*/
static guint update_ctrl_state(DirPane *dp, guint state)
{
	CtrlMode	mode = dp->main->cfg.dp_format[dp->index].ctrl_mode;

	switch(mode)
	{
		case CTM_IGNORE:
			break;
		case CTM_SET:
			state |= GDK_CONTROL_MASK;
			break;
		case CTM_RESET:
			state &= ~GDK_CONTROL_MASK;
			break;
		case CTM_TOGGLE:
			state ^= GDK_CONTROL_MASK;
			break;
	}
	return state;
}

/* 1998-06-18 -	This handles the situation when a user clicks on a row in a directory pane. Might envoke
**		action on files/dirs.
** 1998-06-04 -	Added support for a little menu popping up when the right button is pressed. First it changes
**		to the given pane.
** 1998-06-17 -	Added shift-sensing to extend the current selection. Not so crucial in my mind, since dragging
**		is so supported. But, nice never the less (and somewhat standard, so people might expect it).
** 1998-06-17 -	Added control-sensing to "extend" the current UNselection. Cool?
** 1999-03-05 -	Simplified since we now use GTK+'s built-in CList selection.
** 2003-10-08 -	Implemented Click-M-Click recognition.
*/
static gboolean evt_dirpane_button_press(GtkWidget *wid, GdkEventButton *event, gpointer user)
{
	static DirPane	*last_dp = NULL;
	static GTimeVal	last_time;
	const gchar	*mcmd;
	gint		ret, row, col;
	DirPane		*dp = user;
	gboolean	change = last_dp && dp != last_dp;
	GTimeVal	now;

	dp_path_unfocus(dp);

	/* Handle commands mapped to mouse buttons. This includes the RMB menu, typically. */
	if((event->type == GDK_BUTTON_PRESS) && (mcmd = ctrl_mouse_map(dp->main->cfg.ctrlinfo, event)) != NULL)
	{
		dp_activate_queued(dp);
		last_dp = dp;
		evt_event_set((GdkEvent *) event);
		csq_execute(dp->main, mcmd);
		evt_event_clear(event->type);
		gtk_signal_emit_stop_by_name(GTK_OBJECT(dp->list), "button_press_event");
		return TRUE;
	}

	/* Look for click-m-click, i.e. a rapid click in opposite pane after a selection. */
	g_get_current_time(&now);
	if(change)
	{
		const gchar *cmd = ctrl_clickmclick_get_cmdseq(dp->main->cfg.ctrlinfo);
		if(cmd && *cmd)
		{
			gfloat	elapsed = 1E-6f * (now.tv_usec - last_time.tv_usec) + (now.tv_sec - last_time.tv_sec);
			if(elapsed < ctrl_clickmclick_get_delay(dp->main->cfg.ctrlinfo))
			{
				csq_execute(dp->main, cmd);
				gtk_signal_emit_stop_by_name(GTK_OBJECT(dp->list), "button_press_event");
				return TRUE;
			}
		}
	}

	dp_activate_queued(dp);
	last_dp = dp;
	ret = gtk_clist_get_selection_info(GTK_CLIST(wid), event->x, event->y, &row, &col);
	if(ret && (row < dp->dir.num_rows))
	{
		if((event->type == GDK_2BUTTON_PRESS) && (row == dp->last_row2))
		{
			MainInfo	*min = dp->main;

			dp->dbclk_row = row;

			/* FIXME: This is horribly ugly. If control mode is set to "system default" (i.e., ignore),
			** we need to handle the double-click right now, since we can't punt it to the selection
			** handler as we usually do. This is very ugly, but I think it works.
			*/
			if(min->cfg.dp_format[dp->index].ctrl_mode == CTM_IGNORE)
			{
				doubleclick(dp);
				dp_unselect(dp, row);
			}
			gtk_signal_emit_stop_by_name(GTK_OBJECT(dp->list), "button_press_event");
		}
		else
			event->state = update_ctrl_state(dp, event->state);

		dp->last_row2 = dp->last_row;
		dp->last_row  = row;
		last_time = now;
	}
	else
	{
		dp->last_row = dp->last_row2 = -1;
		last_time.tv_sec = last_time.tv_usec = 0;	/* If click was outside, it can't start c-m-c. */
	}
	return FALSE;
}

static void evt_dirpane_col_click(GtkWidget *wid, gint column, gpointer user)
{
	DirPane	*dp = user;
	DPSort	*sort;

	if(!dp_activate(dp))		/* Only change sort if click didn't cause pane change. */
	{
		DHSel	*sel;

		sort = &dp->main->cfg.dp_format[dp->index].sort;
		if(sort->content == dp->main->cfg.dp_format[dp->index].format[column].content)
			sort->invert ^= 1;	/* Clicking on active title toggles inversion. */
		else
			sort->content = dp->main->cfg.dp_format[dp->index].format[column].content;

		sel = dph_dirsel_new(dp);
		clear_selection_stats(&dp->dir.sel);
		dp_resort(dp);
		dp_redisplay(dp);
		dph_dirsel_apply(dp, sel);
		dph_dirsel_destroy(sel);
	}
}

/* 1999-05-15 -	User resized a column. Update format. I hope this is a new signal... */
static void evt_dirpane_resize_column(GtkWidget *wid, gint column, gint width, gpointer user)
{
	DirPane	*dp = user;

	dp->main->cfg.dp_format[dp->index].format[column].width = width;
	cfg_modified_set(dp->main);
}

/* ----------------------------------------------------------------------------------------- */

/* 1998-05-29 -	Compute a result for qsort() comparison function, taking care to keep
**		directories on top of files at all times.
** 1998-06-07 -	Now implements all the sorting modes offered by Opus, meaning directories
**		can be first, last, or mixed with the files. The inversion is also done here,
**		making it a lot more "true" than the old hack.
** 2000-04-16 -	Now resolves collisions by comparing names. This could be extended in the
**		future to allow user-controlled multi-level sorting.
*/
static gint qsort_result(gint s, const DirRow *da, const DirRow *db)
{
	gint	isda, isdb;

	if(s == 0)		/* Did the two rows compare equal? */
		s = the_strcmp(DP_ROW_NAME(da), DP_ROW_NAME(db));	/* Resolve using names. */

	if(the_sort.invert)
		s = -s;

	if(DP_ROW_GET_STAT(da) != NULL)
		isda = S_ISDIR(DP_ROW_STAT(da).st_mode);
	else
		isda = S_ISDIR(DP_ROW_LSTAT(da).st_mode);

	if(DP_ROW_GET_STAT(db) != NULL)
		isdb = S_ISDIR(DP_ROW_STAT(db).st_mode);
	else
		isdb = S_ISDIR(DP_ROW_LSTAT(db).st_mode);
	switch(the_sort.mode)
	{
		case DPS_DIRS_FIRST:
			if(isda == isdb)
				return s;
			else if(isda)
				return -1;
			return 1;
		case DPS_DIRS_LAST:
			if(isda == isdb)
				return s;
			else if(isdb)
				return -1;
			return 1;
		case DPS_DIRS_MIXED:
			return s;
	}
	return s;		/* I bet this doesn't ever run. Gcc doesn't. */
}

/* 2000-04-16 -	Compare names for sorting. Since name-comparison is now automatically
**		done as a collision-resolver by qsort_result() above, we don't need to
**		do the actual string comparison here anymore.
*/
static gint qsort_cmp_name(gconstpointer ap, gconstpointer bp)
{
	return qsort_result(0, ap, bp);
}

/* 1998-05-29 -	A hairy macro, which can be used to generate sorting data from two dirrows, by numerically
**		comparing one field in the stat() datas. The ARG argument should be the name of the field
**		to compare, e.g. st_size, st_mode etc.
*/
#define	QBODY(ARG)	{const DirRow *da = ap, *db = bp;\
			gint sd;\
			\
			sd = (DP_ROW_LSTAT(da). ARG) < (DP_ROW_LSTAT(db). ARG) ? -1 : (DP_ROW_LSTAT(da). ARG) > (DP_ROW_LSTAT(db). ARG);\
			return qsort_result(sd, da, db);\
			}

/* This does the same as the one above, but is specialized to deal with time_t values, and thus
** goes through the difftime() function to do the comparison.
*/
#define	QTIMEBODY(ARG)	{const DirRow	*da = ap, *db = bp;\
			gint	sd;\
			\
			sd = difftime(DP_ROW_LSTAT(da). ARG, DP_ROW_LSTAT(db). ARG);\
			return qsort_result(sd, da, db);\
			}

static gint qsort_cmp_size(gconstpointer ap, gconstpointer bp)
{
	QBODY(st_size);
}

static gint qsort_cmp_blocks(gconstpointer ap, gconstpointer bp)
{
	QBODY(st_blocks);
}

static gint qsort_cmp_modeoct(gconstpointer ap, gconstpointer bp)
{
	QBODY(st_mode);
}

static gint qsort_cmp_nlink(gconstpointer ap, gconstpointer bp)
{
	QBODY(st_nlink);
}

static gint qsort_cmp_uiddec(gconstpointer ap, gconstpointer bp)
{
	QBODY(st_uid);
}

static gint qsort_cmp_uidstr(gconstpointer ap, gconstpointer bp)
{
	return qsort_result(the_strcmp( usr_lookup_uname(DP_ROW_LSTAT(((DirRow *) ap)).st_uid),
					usr_lookup_uname(DP_ROW_LSTAT(((DirRow *) bp)).st_uid)),
					ap, bp);
}

static gint qsort_cmp_giddec(gconstpointer ap, gconstpointer bp)
{
	QBODY(st_gid);
}

static gint qsort_cmp_gidstr(gconstpointer ap, gconstpointer bp)
{
	return qsort_result(the_strcmp( usr_lookup_gname(DP_ROW_LSTAT(((DirRow *) ap)).st_gid),
					usr_lookup_gname(DP_ROW_LSTAT(((DirRow *) bp)).st_gid)),
					ap, bp);
}

static gint qsort_cmp_dev(gconstpointer ap, gconstpointer bp)
{
	QBODY(st_rdev);
}

static gint qsort_cmp_devmaj(gconstpointer ap, gconstpointer bp)
{
	QBODY(st_rdev >> 8);
}

static gint qsort_cmp_devmin(gconstpointer ap, gconstpointer bp)
{
	QBODY(st_rdev & 0xFF);
}

static gint qsort_cmp_atime(gconstpointer ap, gconstpointer bp)
{
	QTIMEBODY(st_atime);
}

static gint qsort_cmp_mtime(gconstpointer ap, gconstpointer bp)
{
	QTIMEBODY(st_mtime);
}

static gint qsort_cmp_ctime(gconstpointer ap, gconstpointer bp)
{
	QTIMEBODY(st_ctime);
}

/* 2000-04-17 -	Compare two rows based on their type names. */
static gint qsort_cmp_type(gconstpointer ap, gconstpointer bp)
{
	const DirRow	*da = ap, *db = bp;

	return qsort_result(the_strcmp(DP_ROW_TYPE(da)->name, DP_ROW_TYPE(db)->name), ap, bp);
}

/* 1999-06-12 -	Compare two rows based on the styles they use. This is for icon column sorting. */
static gint qsort_cmp_style(gconstpointer ap, gconstpointer bp)
{
	const DirRow	*da = ap, *db = bp;

	return qsort_result(stl_style_compare_hierarchy(DP_ROW_TYPE(da)->style, DP_ROW_TYPE(db)->style), ap, bp);
}

typedef gint	(*sorter)(gconstpointer ap, gconstpointer bp);

/* 1998-05-29 -	Select a sorting function based on the given pane. */
static sorter choose_sorter(DirPane *dp)
{
	switch(dp->main->cfg.dp_format[dp->index].sort.content)
	{
		case DPC_NAME:			return qsort_cmp_name;
		case DPC_SIZE:			return qsort_cmp_size;
		case DPC_BLOCKS:		return qsort_cmp_blocks;
		case DPC_BLOCKSIZE:		return qsort_cmp_blocks;
		case DPC_MODENUM:	/* Fall-through. */
		case DPC_MODESTR:		return qsort_cmp_modeoct;
		case DPC_NLINK:			return qsort_cmp_nlink;
		case DPC_UIDNUM:		return qsort_cmp_uiddec;
		case DPC_UIDSTR:		return qsort_cmp_uidstr;
		case DPC_GIDNUM:		return qsort_cmp_giddec;
		case DPC_GIDSTR:		return qsort_cmp_gidstr;
		case DPC_DEVICE:		return qsort_cmp_dev;
		case DPC_DEVMAJ:		return qsort_cmp_devmaj;
		case DPC_DEVMIN:		return qsort_cmp_devmin;
		case DPC_ATIME:			return qsort_cmp_atime;
		case DPC_MTIME:			return qsort_cmp_mtime;
		case DPC_CTIME:			return qsort_cmp_ctime;
		case DPC_TYPE:			return qsort_cmp_type;
		case DPC_ICON:			return qsort_cmp_style;

		case DPC_NUM_TYPES:
			;
	}
	return NULL;
}

/* 2009-03-05 -	Compare two decimal strings, without converting the text to binary.
 *		Not super-efficient, but should be OK for this application, which
 *		is sorting of directory entries. Sample inputs: "4712" and "42".
*/
static gint compare_decimal(const gchar **a, const gchar **b)
{
	const gchar	*ap = *a, *bp = *b;
	const gchar	*ba, *bb;
	gsize		da, db;

	/* Skip past any leading zeroes, as they are not significant. */
	while(*ap == '0')
		ap++;
	while(*bp == '0')
		bp++;
	/* Save base pointers; this is where the numbers start. */
	ba = ap;
	bb = bp;
	/* Skip to the end of each number. */
	while(isdigit(*ap))
		ap++;
	while(isdigit(*bp))
		bp++;
	/* Now we can compute the number of digits in each. If they differ,
	 * the difference gives the relationship immediately, since there
	 * are no leading zeroes. "99" must be less than "100".
	*/
	da = ap - ba;
	db = bp - bb;
	if(da < db)
		return -1;
	if(db < da)
		return 1;
	/* Answer might be 0, update pointers so caller can continue. */
	*a = ap;
	*b = bp;
	return strncmp(ba, bb, da);
}

/* 1999-08-29 -	Compare two strings, while dealing with embedded numerals in an "intelligent"
**		fashion. Will compare "yo-12" as being greater than "yo-5", for example.
*/
static gint numeric_strcmp(const gchar *a, const gchar *b)
{

	for(;*a && *b;)
	{
		if(isdigit((guchar) *a) && isdigit((guchar) *b))
		{
			gint	rel;

			rel = compare_decimal(&a, &b);
			if(rel != 0)
				return rel;
		}
		else if(*a != *b)
			break;
		else
			a++, b++;
	}
#if defined ENABLE_NLS
	{
		gchar	buf[4] = { 0, 0, 0, 0 };

		buf[0] = *a;
		buf[2] = *b;
		return the_strcmp_case(buf, buf + 2);
	}
#else
	return *a - *b;
#endif
}

/* 1999-08-29 -	Compare two strings, while dealing with embedded numerals in an "intelligent" fashion.
**		Will compare "yo-12" as being greater than "yo-5", for example. Ignores case.
** 2002-06-16 -	Made slightly more I18N-aware, through minimal use of strcoll(). Not Unicode clean!
*/
static gint numeric_strcasecmp(const gchar *a, const gchar *b)
{
	for(;*a && *b;)
	{
		if(isdigit((guchar) *a) && isdigit((guchar) *b))
		{
			gint	rel;

			rel = compare_decimal(&a, &b);
			if(rel != 0)
				return rel;
		}
		else if(tolower((guchar) *a) != tolower((guchar) *b))
			break;
		else
			a++, b++;
	}
#if defined ENABLE_NLS
	{
		gchar	buf[4] = { 0, 0, 0, 0 };

		buf[0] = *a;
		buf[2] = *b;
		return the_strcmp_nocase(buf, buf + 2);
	}
#else
	return tolower((guchar) *a) - tolower((guchar) *b);
#endif
}

/* 1999-04-27 -	Resort the rows of <dp>, according to its format's sort info. Does not cause
**		a redisplay; call dp_redisplay() for that.
** 1999-05-30 -	Fixed a BIG problem: resorting messed up the stat->lstat connections on non-link
**		rows. Very, very bad.
*/
void dp_resort(DirPane *dp)
{
	guint	i;

	/* "NULL out" any non-link stat fields, to avoid breaking links during resorting. */
	for(i = 0; i < dp->dir.num_rows; i++)
	{
		if(DP_ROW_GET_STAT(dp->dir.row + i) == &DP_ROW_LSTAT(dp->dir.row + i))
			DP_ROW_SET_STAT(dp->dir.row + i, NULL);
	}

	the_sort   = dp->main->cfg.dp_format[dp->index].sort;	/* Must be accessible by qsort() callbacks. */
	the_strcmp = the_sort.numerical ? (the_sort.nocase  ? numeric_strcasecmp : numeric_strcmp) :
			the_sort.nocase ? the_strcmp_nocase : the_strcmp_case;

	/* Resort the DirRow structures. */
	qsort(dp->dir.row, dp->dir.num_rows, sizeof *dp->dir.row, choose_sorter(dp));

	/* Now re-connect any NULL stat links, thus making them point at the lstat fields again. */
	for(i = 0; i < dp->dir.num_rows; i++)
	{
		if(DP_ROW_GET_STAT(dp->dir.row + i) == NULL)
			DP_ROW_SET_STAT(dp->dir.row + i, &DP_ROW_LSTAT(dp->dir.row + i));
	}
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-04-27 -	Redisplay given pane. Will totally clear the GtkCList widget, and reformat
**		all row data into it. Does *not* resort the pane's contents; use (the new)
**		dp_resort() function for that. Also does *not* preserve the set of selected
**		rows or the vertical position; use dp_redisplay_preserve() for that.
*/
void dp_redisplay(DirPane *dp)
{
	gchar	buffer[DP_MAX_COLUMNS][2 * PATH_MAX], *buf[DP_MAX_COLUMNS];
	guint	i, j, row, nc;

	dp_freeze(dp);						/* Reduce ugly flicker. */
	gtk_clist_clear(dp->list);
	nc = dp->main->cfg.dp_format[dp->index].num_columns;
	for(i = 0; i < dp->dir.num_rows; i++)
	{
		for(j = 0; j < nc; j++)
			buf[j] = buffer[j];
		for(j = 0; j < nc; j++)
			dpf_format_column(dp->main, dp, &dp->dir.row[i], j, &buf[j]);
		row = gtk_clist_append(dp->list, buf);
		gtk_clist_set_row_data(GTK_CLIST(dp->list), row, &dp->dir.row[i]);
		dpf_post_format_row(dp->main, dp, row);
	}
	dp_thaw(dp);
}

/* 1999-04-27 -	Redisplay given pane, keeping both the selected set and the vertical position.
**		This is the one to use in most cases.
*/
void dp_redisplay_preserve(DirPane *dp)
{
	DHSel	*sel;
	gfloat	vpos;

	sel  = dph_dirsel_new(dp);
	vpos = dph_vpos_get(dp);
	dp_redisplay(dp);
	dph_vpos_set(dp, vpos);
	dph_dirsel_apply(dp, sel);
	dph_dirsel_destroy(sel);
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-10 -	Unfocus the currently focused row of <dp>, if any. */
void dp_unfocus(DirPane *dp)
{
	if((dp != NULL) && (dp->focus_row >= 0))
	{
		gtk_clist_set_row_style(dp->list, dp->focus_row, NULL);
		dp->old_focus_row = dp->focus_row;
		dp->focus_row = -1;
	}
}

/* 1999-05-10 -	Attempt to focus the <row>'th row of <dp>. If there was a previous
**		row focused, unfocus it first.
** 2000-02-03 -	Rewritten, slightly. Now disables focusing if someone attempts to
**		focus outside of the legal range, relying on higher-level code to
**		apply sugar like edge wrapping.
*/
void dp_focus(DirPane *dp, gint row)
{
	if(focus_style == NULL)
	{
		focus_style = gtk_style_copy(gtk_widget_get_style(GTK_WIDGET(dp->list)));

		focus_style->base[GTK_STATE_NORMAL] = dp->main->cfg.dp_colors.color[DPCOL_FOCUS_UNSELECTED];
		focus_style->bg[GTK_STATE_SELECTED] = dp->main->cfg.dp_colors.color[DPCOL_FOCUS_SELECTED];
	}

	dp_unfocus(dp);

	if((row < 0) || (row >= dp->dir.num_rows))
	{
		dp->focus_row = -1;
		return;
	}

	/* Bring the focused row into view, minimizing disruption while doing so. */
	if(gtk_clist_row_is_visible(dp->list, row) != GTK_VISIBILITY_FULL)
	{
		if((dp->focus_row != -1))		/* Is there an already focused row? */
		{
			gint	delta = row - dp->focus_row;

			if(ABS(delta) == 1)
				gtk_clist_moveto(dp->list, row, -1, delta < 0 ? 0.0 : 1.0, -1);
			else
				gtk_clist_moveto(dp->list, row, -1, 0.5, 0.5);
		}
		else
			gtk_clist_moveto(dp->list, row, -1, 0.5, 0.5);	/* If no previous focus, center the new one. */
	}
	gtk_clist_set_row_style(dp->list, row, focus_style);
	dp->focus_row = row;
}

/* 2003-11-30 -	A new dir has been entered, figure out if there should be focusing. */
void dp_focus_enter_dir(DirPane *dp)
{
	if(dp->focus_row != -1)
		return;
	if(dp->old_focus_row == -1)
		return;
	dp_focus(dp, 0);
}

/* ----------------------------------------------------------------------------------------- */

/* 1998-05-25 -	Moved this old workhorse into the new dirpane module, and discovered that
**		it wasn't commented. Lazy me. Also modified its prototype, to take the
**		DirPane to display rather than figuring it out from MainInfo.
*/
void dp_show_stats(DirPane *dp)
{
	MainInfo *min;

	if((min = dp->main) != NULL)
	{
		static gchar	size_format[128] = "";
		gchar		buf[256], selbuf[32] = "", totbuf[32] = "", *ptr = buf;
		const SelInfo	*sel = &dp->dir.sel;

		ptr += g_snprintf(ptr, sizeof buf, _("%u/%u dirs, %u/%u files"),
				  sel->num_dirs, dp->dir.tot_dirs,
				  sel->num_files, dp->dir.tot_files);
		sze_put_size64(selbuf, sizeof selbuf, sel->num_bytes,    SZE_AUTO);
		sze_put_size64(totbuf, sizeof totbuf, dp->dir.tot_bytes, SZE_AUTO);
		if(size_format[0] == '\0')
		{
			/* Format into static variable, lives for the duration of the program. */
			g_snprintf(size_format, sizeof size_format, _(" (%%s/%%s, %%%s/%%%s blocks)"), G_GUINT64_FORMAT, G_GUINT64_FORMAT);
		}
		sprintf(ptr, size_format, selbuf, totbuf, sel->num_blocks, dp->dir.tot_blocks);
		if(dp->dir.fs.valid)
		{
			gchar	dbuf[32], pbuf[32];

			sze_put_size64(dbuf, sizeof dbuf, (guint64) dp->dir.fs.stat.f_bsize * (guint64) (dp->dir.fs.stat.f_blocks - dp->dir.fs.stat.f_bavail), SZE_AUTO);
			g_snprintf(pbuf, sizeof pbuf, "%.1f%%",
				100.0 *
				((gdouble) (dp->dir.fs.stat.f_blocks - dp->dir.fs.stat.f_bavail)
				/
				(gdouble) dp->dir.fs.stat.f_blocks));
			ptr += sprintf(ptr, _(", %s (%s) used"), dbuf, pbuf);
			sze_put_size64(dbuf, sizeof dbuf, (guint64) dp->dir.fs.stat.f_bsize * (guint64) dp->dir.fs.stat.f_bavail, SZE_AUTO);
			ptr += sprintf(ptr, _(", %s free"), dbuf);
		}
		gtk_label_set_text(GTK_LABEL(min->gui->top), buf);
	}
}

/* Blend from <a> to <b>. At <alpha>=0, <dst> is set to <a>, at 1 it is <b>.
** This isn't as classy as a HSL-based blend, but on the other hand it is
** a cute four lines of code.
*/
static void blend(GdkColor *dst, const GdkColor *a, const GdkColor *b, gfloat alpha)
{
	const float ia = 1.0f - alpha;

	dst->red   = alpha * a->red   + ia * b->red;
	dst->green = alpha * a->green + ia * b->green;
	dst->blue  = alpha * a->blue  + ia * b->blue;
}

/* 1998-09-06 -	Set background color of all <dp>'s column buttons. If <col> is NULL,
**		the default background color (actually, the entire default style) will be used.
*/
static void dp_set_col_color(DirPane *dp, const GdkColor *col)
{
	guint	i;

	if(dp != NULL)
	{
		GdkColor	bright, white, *pl = NULL;

		if(col != NULL)
		{
			white.red = white.green = white.blue = ~0;
			blend(&bright, &white, col, 0.7f);
			pl = &bright;
		}
		for(i = 0; i < GTK_CLIST(dp->list)->columns; i++)
		{
			gtk_widget_modify_bg(GTK_CLIST(dp->list)->column[i].button, GTK_STATE_NORMAL,   col);
			gtk_widget_modify_bg(GTK_CLIST(dp->list)->column[i].button, GTK_STATE_PRELIGHT, pl);
		}
	}
}

/* 2002-07-14 -	Do the pane rendering necessary to change active pane from <from> into <to>. */
static void activate_render(DirPane *from, DirPane *to)
{
	if(from != NULL)
	{
		dp_unfocus(from);
		dp_set_col_color(from, NULL);
	}
	if(to != NULL)
	{
		dp_set_col_color(to, &to->main->cfg.dp_colors.color[DPCOL_PANE_SELECTED]);
		dp_show_stats(to);
		if(to->old_focus_row >= 0)
			dp_focus(to, to->old_focus_row);
	}
}

/* 1999-06-09 -	Activate <dp>, making it the source pane for all operations. Returns TRUE if the activation
**		meant that another pane was deactivated, FALSE if <dp> was already the activate pane.
*/
gboolean dp_activate(DirPane *dp)
{
	if(dp != NULL && dp->main->gui->cur_pane != dp)
	{
		activate_render(dp->main->gui->cur_pane, dp);
		dp->main->gui->cur_pane = dp;
		return TRUE;
	}
	return FALSE;
}

/* 2002-07-13 -	This runs when GTK+ is idle, and does the pane activation rendering, once. */
static gint idle_activate(gpointer data)
{
	activate_render(the_activate_queue_info.old_cur, ((MainInfo *) data)->gui->cur_pane);
	gtk_idle_remove(the_activate_queue_info.handler);
	the_activate_queue_info.handler = 0U;
	return 0;
}

/* 2002-07-13 -	Activate a pane, but queue the rendering until GTK+ is idle. This works around
**		a very annoying bug/misfeature which causes "ghost" pane scrolling to occur.
**		We only queue the rendering, since that seems to suffice, and we need to really
**		change gentoo's idea of "active pane" immediately or things break.
*/
static void dp_activate_queued(DirPane *dp)
{
	if(the_activate_queue_info.handler)
		gtk_idle_remove(the_activate_queue_info.handler);
	the_activate_queue_info.handler = gtk_idle_add(idle_activate, dp->main);
	the_activate_queue_info.old_cur = dp->main->gui->cur_pane;
	dp->main->gui->cur_pane = dp;
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-03-05 -	Answer whether given pane has one or more rows selected or not. */
gboolean dp_has_selection(DirPane *dp)
{
	return (dp->dbclk_row != -1) || (dp->focus_row != -1) || (GTK_CLIST(dp->list)->selection != NULL && !no_repeat);
}

/* 1999-03-06 -	Answer whether given <row> is currently selected. A bit more efficient
**		than getting the entire list through dp_get_selection(), but only if
**		you're not going to iterate the selection anyway.
*/
gboolean dp_is_selected(DirPane *dp, DirRow *row)
{
	if((dp != NULL) && (row != NULL))
	{
		if((GTK_CLIST(dp->list)->selection == NULL) && dpf_get_fake_select() && (DP_ROW_INDEX(dp, row) == dp->focus_row))
			return TRUE;
		return g_list_find(GTK_CLIST(dp->list)->selection, GINT_TO_POINTER(DP_ROW_INDEX(dp, row))) != NULL;
	}
	return FALSE;
}

/* 1999-03-15 -	Compare two rows based on their indexes in the directory. Assumes their addresses
**		are comparable, which they really should be.
*/
static gint compare_rows(gconstpointer a, gconstpointer b)
{
	return GPOINTER_TO_UINT(a) - GPOINTER_TO_UINT(b);
}

/* 1999-03-04 -	Return a GSList of selected rows. Each item's data member is a DirRow.
**		Knows how to deal with a double click, too. This is going to be the new
**		interface for all commands.
** 1999-03-15 -	Now sorts the rows in address order, since otherwise the selection retains
**		the order in which it was done by the user, and that is simply confusing.
*/
GSList * dp_get_selection(DirPane *dp)
{
	GList	*iter;
	GSList	*sel = NULL;

	no_repeat = FALSE;

	if(dp->dbclk_row != -1)
		sel = g_slist_append(sel, gtk_clist_get_row_data(GTK_CLIST(dp->list), dp->dbclk_row));
	else if((dp->focus_row != -1) && dpf_get_fake_select() && (GTK_CLIST(dp->list)->selection == NULL))
	{
		sel = g_slist_append(sel, gtk_clist_get_row_data(GTK_CLIST(dp->list), dp->focus_row));
		no_repeat = TRUE;
	}
	else
		for(iter = GTK_CLIST(dp->list)->selection; iter != NULL; iter = g_list_next(iter))
			sel = g_slist_insert_sorted(sel, gtk_clist_get_row_data(GTK_CLIST(dp->list), GPOINTER_TO_INT(iter->data)),
							compare_rows);
	if(sel)
		fam_rescan_block();

	return sel;
}

/* 1999-03-06 -	Get the "full" selection, regardless of whether there's a double clicked row or not.
**		This should only be used if you really know what you're doing, and never by actual
**		commands (which need the double-click support).
*/
GSList * dp_get_selection_full(const DirPane *dp)
{
	const GList	*iter;
	GSList		*sel = NULL;

	for(iter = GTK_CLIST(dp->list)->selection; iter != NULL; iter = g_list_next(iter))
		sel = g_slist_append(sel, gtk_clist_get_row_data(GTK_CLIST(dp->list), GPOINTER_TO_INT(iter->data)));
	if(sel)
		fam_rescan_block();
	return sel;
}

#if 0
/* 1999-03-05 -	A little routine handy for debugging. */
void dp_print_selection(DirPane *dp, const gchar *tag)
{
	GList	*iter;

	fprintf(stderr, "[%10s] Selection: ", tag);
	for(iter = GTK_CLIST(dp->list)->selection; iter != NULL; iter = g_list_next(iter))
		fprintf(stderr, "'%d' ", GPOINTER_TO_INT(iter->data));
	fprintf(stderr, "\n");
}
#endif

/* 1999-03-04 -	Free a selection list, handy when you're done traversing it. */
void dp_free_selection(GSList *sel)
{
	if(sel != NULL)
	{
		g_slist_free(sel);
		fam_rescan_unblock();
	}
}

/* ----------------------------------------------------------------------------------------- */

/* 1998-05-25 -	Count entries as they are read in. */
static void statistics_add_entry(DirPane *dp, struct stat *stat)
{
	if(S_ISDIR(stat->st_mode))
		dp->dir.tot_dirs++;
	else
		dp->dir.tot_files++;
	dp->dir.tot_bytes  += stat->st_size;
	dp->dir.tot_blocks += stat->st_blocks;
}

/* 1999-04-08 -	Clear the selected statistics. */
static void clear_selection_stats(SelInfo *sel)
{
	if(sel != NULL)
	{
		sel->num_dirs = 0;
		sel->num_files = 0;
		sel->num_bytes = 0;
		sel->num_blocks = 0;
	}
}

/* 1999-04-09 -	Clear the total statistics fields for <dp>. */
static void clear_total_stats(DirPane *dp)
{
	dp->dir.tot_files = dp->dir.tot_dirs = 0;
	dp->dir.tot_bytes = dp->dir.tot_blocks = 0;
 	dp->dir.fs.valid = FALSE;
}

/* 1999-01-03 -	Clear the statistics for <dp>. Useful when the pane's path is about to change. */
static void clear_stats(DirPane *dp)
{
	clear_total_stats(dp);
	clear_selection_stats(&dp->dir.sel);
	dp->last_row = dp->last_row2 = -1;
}

/* 1999-04-08 -	Update the selection statistics for <dp>, by simply flushing them and recomputing from
**		scratch. Does NOT use dp_get_selection(), for performance reasons (this is run on every
**		unselection).
*/
static void update_selection_stats(DirPane *dp)
{
	GList	*iter;
	DirRow	*dr;
	SelInfo	*sel = &dp->dir.sel;

	clear_selection_stats(sel);
	for(iter = GTK_CLIST(dp->list)->selection; iter != NULL; iter = g_list_next(iter))
	{
		dr = gtk_clist_get_row_data(GTK_CLIST(dp->list), GPOINTER_TO_INT(iter->data));
		if(S_ISDIR(DP_ROW_LSTAT(dr).st_mode))
			sel->num_dirs++;
		else
			sel->num_files++;
		sel->num_bytes  += DP_ROW_LSTAT(dr).st_size;
		sel->num_blocks += DP_ROW_LSTAT(dr).st_blocks;
	}
}

/* 1999-04-09 -	Update the tot_XXX fields in <dp>'s directory statistics. Handy after a
**		(Get|Clear)Size.
*/
void dp_update_stats(DirPane *dp)
{
	DirRow	*dr;
	guint	i;

	clear_total_stats(dp);
	clear_selection_stats(&dp->dir.sel);
	for(i = 0; i < dp->dir.num_rows; i++)
	{
		dr = &dp->dir.row[i];

		if(S_ISDIR(DP_ROW_LSTAT(dr).st_mode))
			dp->dir.tot_dirs++;
		else
			dp->dir.tot_files++;
		dp->dir.tot_bytes  += DP_ROW_LSTAT(dr).st_size;
		dp->dir.tot_blocks += DP_ROW_LSTAT(dr).st_blocks;
	}
}

/* 1999-03-30 -	Update information about filesystem for given <dp>. */
static void update_fs_info(DirPane *dp)
{
	gint	old_errno = errno;

#if defined HAVE_STRUCT_STATVFS
	dp->dir.fs.valid = statvfs(dp->dir.path, &dp->dir.fs.stat) == 0;
#else
	dp->dir.fs.valid = statfs(dp->dir.path, &dp->dir.fs.stat) == 0;
#endif
	errno = old_errno;
}

/* 2003-11-11 -	Grow allocation from <n>. Grows expontentially up to a limit, then linearly. Heuristic. */
static gsize row_grow(gsize n)
{
	const gsize	LINEAR = 128;

	if(n == 0)
		return 1;
	else if(n < LINEAR)
		return n << 1;
	return n + LINEAR;
}

/* 2003-11-11 -	Store a row into given pane. Will grow buffers if required. Gruesome. */
static void store_row(DirPane *dp, const char *name, const struct stat *stbuf, const FType *unknown)
{
	DirContents	*dir = &dp->dir;
	DirRow		*row;
	gsize		flen, dlen, nsize;
	gchar		*dpl;
	GError		*err = NULL;
	int		errbuf = errno;

	/* Enough DirRow structures allocated? */
	if(dir->num_rows == dir->row_alloc)
	{
		dir->row_alloc = row_grow(dir->row_alloc);
		dir->row = g_realloc(dir->row, dir->row_alloc * sizeof *dir->row);
	}

	flen = strlen(name);
	nsize = 0;
	/* This call sometimes contaminates errno even when successful. */
	dpl = g_filename_to_utf8(name, flen, NULL, &dlen, &err);
	if(dpl != NULL)
	{
		dlen += 1;
		nsize += dlen;
		errno = errbuf;	/* If all went well, restore errno. */
	}
	else
	{
		printf("Unable to convert '%s' to UTF-8, check G_FILENAME_ENCODING\n", name);
		if(err != NULL)
		{
/*			err_printf(dp->min, _("Conversion of filename to UTF-8 failed, code %d (%s)"), err->code, err->message);
			err_show(dp->min);
*/
 /*			printf(" glib said code %d, '%s'\n", err->code, err->message);*/
			g_error_free(err);
		}
		dpl = g_strdup(_("!!Invalid filename!!"));
		dlen = g_utf8_strlen(dpl, -1) + 1;
	}

	flen += 1;
	nsize += flen;
/*	printf("'%s' is %u, plus %u bytes of UTF-8, total %u (%u with margin) [%u/%u bytes used]\n", name, flen, dlen, nsize, nsize + PATH_MAX + 256,
	       dir->name_use, dir->name_alloc);
*/
	/* Enough name space allocated? */
	if((dir->name_alloc - dir->name_use) < nsize + PATH_MAX)
	{
		dir->name_alloc += PATH_MAX + nsize + 256;	/* Enough for this name, even if symlink, and a few more. */
		dir->name = g_realloc(dir->name, dir->name_alloc);
	}
	row = dir->row + dir->num_rows;
/*	printf("writing %u bytes of disk name at %u\n", flen, dir->name_use);*/
	memcpy(dir->name + dir->name_use, name, flen);
	DP_ROW_NAME(row) = GUINT_TO_POINTER(dir->name_use);	/* Store offset only, since we might realloc() during read. */
	dir->name_use += flen;
/*	printf(" writing %u bytes of display name at %u\n", dlen, dir->name_use);*/
	memcpy(dir->name + dir->name_use, dpl, dlen);
	g_free(dpl);
	DP_ROW_NAMED(row) = GUINT_TO_POINTER(dir->name_use);
	dir->name_use += dlen;

	DP_ROW_LSTAT(row) = *stbuf;
	DP_ROW_SET_STAT(row, NULL);
	DP_ROW_SET_STAT(row, &DP_ROW_LSTAT(row));	/* GTK2: This sometimes works. */
	DP_ROW_FLAGS(row) = 0;
	DP_ROW_TYPE(row) = unknown;
	if(!S_ISDIR(DP_ROW_LSTAT(row).st_mode))
		DP_ROW_FLAGS_SET(row, DPRF_HAS_SIZE);

	DP_ROW_SET_LINKNAME(row, NULL);

	if(S_ISLNK(DP_ROW_LSTAT(row).st_mode))	/* Is it a symlink? */
	{
		gint	len, oen = errno;

		len = readlink(name, dir->name + dir->name_use, PATH_MAX);
		if(len < 0)
			len = 0;
		dir->name[dir->name_use + len] = '\0';
		DP_ROW_SET_LINKNAME(row, GUINT_TO_POINTER(dir->name_use));
		/* Do we have enough stat buffer space? */
		if(dir->stat_use >= dir->stat_alloc)
		{
			dir->stat_alloc = row_grow(dir->stat_alloc);
			dir->stat = g_realloc(dir->stat, dir->stat_alloc * sizeof *dir->stat);
		}
		if(stat(dir->name + dir->name_use, dir->stat + dir->stat_use) == 0)
		{
			DP_ROW_SET_STAT(row, GINT_TO_POINTER(dir->stat_use));
			DP_ROW_FLAGS_SET(row, DPRF_LINK_EXISTS);
			dir->stat_use++;
		}
		errno = oen;
		/* Now, we need to figure out the display version of the symlink contents, and store that. */
		dpl = g_filename_to_utf8(dir->name + dir->name_use, len, NULL, &dlen, NULL);
/*		printf("link '%s' converted to display '%s'\n", dir->name + dir->name_use, dpl);*/
		/* Space enough to store the display link name? */
		if(dir->name_alloc - dir->name_use < dlen + 1)
		{
			dir->name_alloc += dlen + 256;
			dir->name = g_realloc(dir->name, dir->name_alloc);
		}
		memcpy(dir->name + dir->name_use + len + 1, dpl, dlen + 1);
		DP_ROW_SET_LINKNAMED(row, GUINT_TO_POINTER(dir->name_use + len + 1));
		g_free(dpl);
		dir->name_use += (len + 1) + (dlen + 1);
	}
	dir->num_rows++;
}

void dp_path_clear(DirPane *dp)
{
	GtkListStore	*store = GTK_LIST_STORE(gtk_combo_box_get_model(GTK_COMBO_BOX(dp->path)));
	gtk_list_store_clear(store);
}

void dp_path_focus(DirPane *dp)
{
	gtk_widget_grab_focus(dp->path);
}

void dp_path_unfocus(DirPane *dp)
{
	gtk_widget_grab_focus(dp->hide);
#if 0
	GdkEvent	fe;

	/* HACK: This fakes a focus-out event on the path entry combo widget, which seems to work. */
	fe.focus_change.type = GDK_FOCUS_CHANGE;
	/* FIXME: If we required GTK+ 2.14 and greater, we could use gtk_widget_get_window(). */
	fe.focus_change.window = GTK_WIDGET(dp->path)->window;
	fe.focus_change.send_event = TRUE;
	fe.focus_change.in = FALSE;
	gtk_main_do_event(&fe);
#endif
}

/* 2003-11-14 -	Brand new version of this very old workhorse. Now reads in the directory in a single
**		pass, using realloc() to grow various buffers on the fly. Slightly more wasteful of
**		memory than the old two-pass version, but less code, more robust, and possibly a
**		little bit faster. Returns TRUE on success.
*/
gboolean dp_enter_dir(DirPane *dp, const gchar *path)
{
	gchar	new_path[PATH_MAX];
	DIR	*dir;
	FType	*unknown;

	if((path[0] != '\0') && (path[0] != G_DIR_SEPARATOR))
		g_snprintf(new_path, sizeof new_path, "%s%s%s", dp->dir.path, G_DIR_SEPARATOR_S, path);
	else
		stu_strncpy(new_path, path, sizeof new_path);

	mnt_entering(dp->main, new_path, MNT_ALWAYS);

	/* To read a directory easily, we need to change to it. */
	if(!fut_cd(new_path, NULL, 0))
	{
		if(dp->dir.path[0] == '\0' && fut_cd(".", NULL, 0))	/* If no previous, try current. */
			return dp_enter_dir(dp, ".");
		return FALSE;
	}

	clear_stats(dp);
	update_fs_info(dp);
	dp_unfocus(dp);

	completion_clear(dp);

	/* Reset buffer usage, reusing existing space if any. */
	dp->dir.num_rows = 0;
	dp->dir.name_use = 0;
	dp->dir.stat_use = 0;

	unknown = typ_type_get_unknown(&dp->main->cfg);

	/* Open directory, and read in all entries. */
	if((dir = opendir(".")) != NULL)
	{
		struct dirent	*de;
		guint		i;
/*		GTimer		*timer;

		timer = g_timer_new();
		g_timer_start(timer);
*/
		/* Read in all items we can find, and let buffers grow as needed. */
		while((de = readdir(dir)) != NULL)
		{
			struct stat	stbuf;

			if(!(dp->main->cfg.dir_filter(de->d_name) && fut_check_hide(dp, de->d_name)))
				continue;
			if(lstat(de->d_name, &stbuf) == 0)
				store_row(dp, de->d_name, &stbuf, unknown);
		}
		closedir(dir);

		/* Now when scan is done, add buffer base to offsets stored, leaving final pointers. */
		typ_identify_begin(dp->main);
		for(i = 0; i < dp->dir.num_rows; i++)
		{
			DirRow	*row = dp->dir.row + i;

/*			printf("%3u: %u -> ", i, GPOINTER_TO_UINT(DP_ROW_NAME(row)));*/
			DP_ROW_NAME(row) = dp->dir.name + GPOINTER_TO_UINT(DP_ROW_NAME(row));
/*			printf("'%s'\n", DP_ROW_NAME(row));
			printf("     %u -> ", GPOINTER_TO_UINT(DP_ROW_NAMED(row)));
*/			DP_ROW_NAMED(row) = dp->dir.name + GPOINTER_TO_UINT(DP_ROW_NAMED(row));
/*			printf("'%s'\n", DP_ROW_NAMED(row));*/
			if(DP_ROW_LINKNAME(row))
			{
				DP_ROW_SET_LINKNAME(row, dp->dir.name + GPOINTER_TO_UINT(DP_ROW_LINKNAME(row)));
				DP_ROW_SET_LINKNAMED(row, dp->dir.name + GPOINTER_TO_UINT(DP_ROW_LINKNAMED(row)));
				if(DP_ROW_FLAGS_CHK(row, DPRF_LINK_EXISTS))
					DP_ROW_SET_STAT(row, dp->dir.stat + GPOINTER_TO_UINT(DP_ROW_GET_STAT(row)));
				else
					DP_ROW_SET_STAT(row, &DP_ROW_LSTAT(row));
			}
			else
				DP_ROW_SET_STAT(row, &DP_ROW_LSTAT(row));
			typ_identify(dp->main, row);
			statistics_add_entry(dp, &DP_ROW_LSTAT(row));
		}
		typ_identify_end(dp->main, dp->dir.path);
/*		g_timer_stop(timer);*/
/*		printf("directory read, uses %d/%d bytes (%.1f%%) of name buffer space for %d rows\n",
			dp->dir.name_use, dp->dir.name_alloc,
			100.0 * dp->dir.name_use / dp->dir.name_alloc,
			dp->dir.num_rows);
*/
/*		printf(" scan completed in %.1f ms\n", 1e3 * g_timer_elapsed(timer, NULL));
		g_timer_destroy(timer);
*/		/* Finally, store the new path in the DirPane. */
		g_strlcpy(dp->dir.path, path, sizeof dp->dir.path);
	}
	dp_resort(dp);
	return TRUE;
}

/* ----------------------------------------------------------------------------------------- */

/* 2002-07-20 -	Rescan contents of given pane.
** NOTE NOTE	This routine does NOT change the error status, i.e. it does not call anything
**		in the error module. This is important, since it's used in the exiting code
**		of generic commands, for instance.
** 2003-09-26 -	Don't activate <dp> for refresh, just do it anyway. Faster, hopefully safe.
*/
void dp_rescan(DirPane *dp)
{
	dph_state_save(dp);
	dp_freeze(dp);
	dp_enter_dir(dp, dp->dir.path);
	dp_redisplay(dp);
	dph_state_restore(dp);
	dp_thaw(dp);
}

/* 2002-07-23 -	Special "weaker" version of dp_rescan(), with the added semantic difference that
**		this is only ever called as part of the "clean-up" *after* a command has finished.
**		The intent is to catch any changes to the pane done by the command, such as a
**		deleted, renamed, moved or otherwise changed file. The reason it has a separate
**		entrypoint is that with FAM, this needs to be stopped.
*/
void dp_rescan_post_cmd(DirPane *dp)
{
#if defined HAVE_FAM
	/* If FAM is compiled in, check if it's working, and if so trust it to handle rescan. */
	if(fam_is_active())
		return;
#endif
	/* If built without FAM, or FAM is not working, just do the rescan (fall-through). */
	dp_rescan(dp);
}

/* ----------------------------------------------------------------------------------------- */

/* 2008-09-20 -	Sets the history strings, available in the combo box. This is to prevent
 *		other modules (dirhistory, I'm looking at you!) from poking the widgets
 *		directly. Abstract, hide, and so on.
*/
void dp_history_set(DirPane *dp, const GList *strings)
{
	GtkListStore	*store = GTK_LIST_STORE(gtk_combo_box_get_model(GTK_COMBO_BOX(dp->path)));
	GtkTreeIter	iter;
	guint		i;

	gtk_list_store_clear(store);
	for(i = 0; strings != NULL; strings = g_list_next(strings))
	{
		const gchar	*str = strings->data;
		if(*str)
		{
			gtk_list_store_append(store, &iter);
			gtk_list_store_set(store, &iter, 0, str, -1);
			i++;
		}
	}
	if(i > 0)
	{
		/* Was there any history set? If so, also jump to the first entry. */
		gtk_signal_handler_block(GTK_OBJECT(gtk_bin_get_child(GTK_BIN(dp->path))), dp->sig_path_activate);
		gtk_signal_handler_block(GTK_OBJECT(dp->path), dp->sig_path_changed);
		gtk_combo_box_set_active(GTK_COMBO_BOX(dp->path), 0);
		gtk_signal_handler_unblock(GTK_OBJECT(dp->path), dp->sig_path_changed);
		gtk_signal_handler_unblock(GTK_OBJECT(gtk_bin_get_child(GTK_BIN(dp->path))), dp->sig_path_activate);
	}
}

/* ----------------------------------------------------------------------------------------- */

/* 2009-07-04 -	Refreshes the split, making sure it's correctly sized according to current settings. */
void dp_split_refresh(MainInfo *min)
{
	gint	pw, ph, np, pos;

	if(min == NULL || min->gui == NULL || min->gui->panes == NULL)
		return;
	if(!GTK_WIDGET_REALIZED(min->gui->panes))
		return;

	/* Inspect the size of the GtkPaned that holds the panes; the window's size
	** doesn't work for vertical due to buttons. This is slightly hackish.
	*/
	gdk_window_get_size(min->gui->panes->window, &pw, &ph);
	np = min->cfg.dp_paning.orientation == DPORIENT_HORIZ ? pw : ph;

	/* Compute the proper new position, depending on the active paning mode. */
	switch(min->cfg.dp_paning.mode)
	{
		case DPSPLIT_FREE:
			return;
		case DPSPLIT_RATIO:
			pos = np * min->cfg.dp_paning.value;
			break;
		case DPSPLIT_ABS_LEFT:
			pos = min->cfg.dp_paning.value;
			break;
		case DPSPLIT_ABS_RIGHT:
			pos = np - min->cfg.dp_paning.value;
			break;
	}
	if(pos > 0)
	{
		g_signal_handler_block(G_OBJECT(min->gui->window), min->gui->sig_main_configure);
		g_signal_handler_block(G_OBJECT(min->gui->panes), min->gui->sig_pane_notify);
		gtk_paned_set_position(GTK_PANED(min->gui->panes), pos);
		g_signal_handler_unblock(G_OBJECT(min->gui->panes), min->gui->sig_pane_notify);
		g_signal_handler_unblock(G_OBJECT(min->gui->window), min->gui->sig_main_configure);
	}
}

/* ----------------------------------------------------------------------------------------- */

/* 2003-11-08 -	Add a pathwidgtry user. Indirect, just adds a function that is later called
**		(in dp_build()) to construct the widgetry. Returns key to use with show().
*/
guint dp_pathwidgetry_add(PageBuilder func)
{
	if(func && !g_slist_find(pathwidgetry_builders, func))
	{
		pathwidgetry_builders = g_slist_append(pathwidgetry_builders, func);
		return g_slist_length(pathwidgetry_builders);
	}
	return 0;
}

/* 2003-11-08 -	Change to a new page of dirpane path widgetry. Returns page. */
GtkWidget ** dp_pathwidgetry_show(DirPane *dp, guint key)
{
	GtkWidget	*page;

	if((page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(dp->notebook), key)) != NULL)
	{
		gtk_notebook_set_page(GTK_NOTEBOOK(dp->notebook), key);
		return gtk_object_get_data(GTK_OBJECT(page), "dp-pathwidgetry");
	}
	return NULL;
}

/* ----------------------------------------------------------------------------------------- */

/* 1998-06-26 -	A lot of code broken out from the giant function below and put here. This
**		makes rebuilding just the list (and not the path entry, for example) a
**		whole lot easier. Useful when config changes...
** 1998-10-26 -	Since the path entry's location now is configurable, this routine is
**		actually never called directly any longer. The true dp_build() always does
**		the job. Function made static.
*/
static void dp_build_list(DPFormat *fmt, DirPane *dp)
{
	GtkWidget	*list;
	gchar		*title[DP_MAX_COLUMNS], name[32];
	guint		i;

	for(i = 0; i < fmt->num_columns; i++)
		title[i] = fmt->format[i].title;
	list = gtk_clist_new_with_titles(fmt->num_columns, title);
	g_snprintf(name, sizeof name, "mainPane%d", dp->index);
	gtk_widget_set_name(list, name);
	GTK_WIDGET_UNSET_FLAGS(list, GTK_CAN_FOCUS);
	for(i = 0; i < fmt->num_columns; i++)	/* A rather gross hack. */
	{
		gtk_clist_set_column_width(GTK_CLIST(list), i, fmt->format[i].width);
		gtk_clist_set_column_justification(GTK_CLIST(list), i, fmt->format[i].just);
		GTK_WIDGET_UNSET_FLAGS(GTK_CLIST(list)->column[i].button, GTK_CAN_FOCUS);
	}
	g_signal_connect(G_OBJECT(list), "button_press_event", G_CALLBACK(evt_dirpane_button_press), dp);
	g_signal_connect(G_OBJECT(list), "click_column", G_CALLBACK(evt_dirpane_col_click), dp);
	g_signal_connect(G_OBJECT(list), "select_row", G_CALLBACK(evt_dirpane_select_row), dp);
	g_signal_connect(G_OBJECT(list), "unselect_row", G_CALLBACK(evt_dirpane_unselect_row), dp);
	g_signal_connect(G_OBJECT(list), "resize_column", G_CALLBACK(evt_dirpane_resize_column), dp);

	gtk_widget_set_events(list, GDK_POINTER_MOTION_MASK);
	if(fmt->scrollbar_always)
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(dp->scwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	else
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(dp->scwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_clist_set_selection_mode(GTK_CLIST(list), GTK_SELECTION_EXTENDED);
	dp->list = GTK_CLIST(list);
}

/* 1998-05-23 -	Rewrote this one, now encloses stuff in a frame, which provides a fairly nice way to
**		indicate current directory. Now also uses the new dirpane formatting/config stuff.
** 1998-06-26 -	Cut away the code above (dp_build_list), making this function a lot leaner.
** 1998-09-06 -	Frame removed, since I've become cool enough to use styles to just change the
**		background color of the pane's column buttons. Looks great!
** 1998-10-26 -	Now supports configuring the position of the path entry (above or below).
*/
GtkWidget * dp_build(MainInfo *min, DPFormat *fmt, DirPane *dp)
{
	GdkPixmap	*pix;
	GdkBitmap	*msk;
	GtkWidget	*hbox, *pr, *ihbox;
	GtkListStore	*model;
	GSList		*iter;

	if(pane_selected != NULL)
	{
		gtk_style_unref(pane_selected);
		pane_selected = NULL;
	}
	if(focus_style != NULL)
	{
		gtk_style_unref(focus_style);
		focus_style = NULL;
	}
	if(dp == NULL)
		return NULL;

	dp->vbox = gtk_vbox_new(FALSE, 0);
	/* "Internal" hbox, holds the scrolled window of the pane and the optional
	** "huge" parent button. If the latter is disabled, it's redundant, but hey.
	*/
	ihbox = gtk_hbox_new(FALSE, 0);

	dp->scwin = gtk_scrolled_window_new(NULL, NULL);
	if(fmt->sbar_pos == SBP_LEFT)
		gtk_scrolled_window_set_placement(GTK_SCROLLED_WINDOW(dp->scwin), GTK_CORNER_TOP_RIGHT);
	else if(fmt->sbar_pos == SBP_RIGHT)
		gtk_scrolled_window_set_placement(GTK_SCROLLED_WINDOW(dp->scwin), GTK_CORNER_TOP_LEFT);
	dp_build_list(fmt, dp);
	gtk_container_add(GTK_CONTAINER(dp->scwin), GTK_WIDGET(dp->list));

	dp->notebook = gtk_notebook_new();
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(dp->notebook), FALSE);
	gtk_notebook_set_show_border(GTK_NOTEBOOK(dp->notebook), FALSE);
	hbox = gtk_hbox_new(FALSE, 0);
	pr   = gtk_button_new();	/* GTK2: This should probably be a stock image. */
	if((pix = gdk_pixmap_colormap_create_from_xpm_d(NULL, gtk_widget_get_default_colormap(), &msk, NULL, icon_dirparent_xpm)) != NULL)
	{
		GtkWidget	*pmap;

		if((pmap = gtk_pixmap_new(pix, msk)) != NULL)
			gtk_container_add(GTK_CONTAINER(pr), pmap);
	}
	GTK_WIDGET_UNSET_FLAGS(pr, GTK_CAN_FOCUS);
	g_signal_connect(G_OBJECT(pr), "clicked", G_CALLBACK(evt_parent_clicked), dp);
	gtk_box_pack_start(GTK_BOX(hbox), pr, FALSE, FALSE, 0);
	gtk_tooltips_set_tip(min->gui->main_tt, pr, _("Move up to the parent directory"), NULL);
	dp->menu_top	 = NULL;
	dp->menu_action  = NULL;
	dp->mitem_action = NULL;

	model = gtk_list_store_new(1, G_TYPE_STRING);
	dp->path = gtk_combo_box_entry_new_with_model(GTK_TREE_MODEL(model), 0);

	dp->complete.compl = gtk_entry_completion_new();
	model = gtk_list_store_new(1, G_TYPE_STRING);
	gtk_entry_completion_set_model(dp->complete.compl, GTK_TREE_MODEL(model));
	gtk_entry_completion_set_text_column(dp->complete.compl, 0);
	gtk_entry_completion_set_inline_completion(dp->complete.compl, TRUE);
	gtk_entry_completion_set_popup_single_match(dp->complete.compl, FALSE);
	gtk_entry_completion_set_match_func(dp->complete.compl, completion_match_function, dp, NULL);
	gtk_entry_set_completion(DP_ENTRY(dp), dp->complete.compl);
	g_signal_connect(G_OBJECT(gtk_bin_get_child(GTK_BIN(dp->path))), "focus_in_event", G_CALLBACK(evt_path_focus), dp);
	g_signal_connect(G_OBJECT(gtk_bin_get_child(GTK_BIN(dp->path))), "focus_out_event", G_CALLBACK(evt_path_unfocus), dp);
	dp->sig_path_activate = g_signal_connect(G_OBJECT(gtk_bin_get_child(GTK_BIN(dp->path))), "activate", G_CALLBACK(evt_path_new), dp);
	dp->sig_path_changed = g_signal_connect(G_OBJECT(dp->path), "changed", G_CALLBACK(evt_path_changed), dp);

	gtk_box_pack_start(GTK_BOX(hbox), dp->path, TRUE, TRUE, 0);
	gtk_tooltips_set_tip(min->gui->main_tt, dp->path, _("Enter path, then press Return to go there"), NULL);
	dp->hide = gtk_toggle_button_new_with_label(_("H"));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(dp->hide), fmt->hide_allowed);
	g_signal_connect(G_OBJECT(dp->hide), "clicked", G_CALLBACK(evt_hide_clicked), dp);
	gtk_box_pack_start(GTK_BOX(hbox), dp->hide, FALSE, FALSE, 0);
	gtk_tooltips_set_tip(min->gui->main_tt, dp->hide, _("Click to enable/disable Hide rule (When pressed in, the hide rule is active, and matching entries are hidden)"), NULL);
	gtk_notebook_append_page(GTK_NOTEBOOK(dp->notebook), hbox, NULL);

	/* If we have any registered pathwidgetry builders, go through and add their output to notebook. */
	for(iter = pathwidgetry_builders; iter != NULL; iter = g_slist_next(iter))
	{
		PageBuilder	func = iter->data;
		GtkWidget	**page;

		if(func && (page = func(min)) != NULL)
		{
			gtk_object_set_data(GTK_OBJECT(*page), "dp-pathwidgetry", page);
			gtk_notebook_append_page(GTK_NOTEBOOK(dp->notebook), *page, NULL);
		}
	}
	if(fmt->huge_parent)
	{
		GtkWidget	*pbtn;

		pbtn = gtk_button_new();
		gtk_button_set_relief(GTK_BUTTON(pbtn), GTK_RELIEF_NONE);
		g_signal_connect(G_OBJECT(pbtn), "clicked", G_CALLBACK(evt_hugeparent_clicked), dp);
		gtk_tooltips_set_tip(min->gui->main_tt, pbtn, _("Move up to the parent directory"), NULL);
		GTK_WIDGET_UNSET_FLAGS(pbtn, GTK_CAN_FOCUS);
		if(dp->index == 0)
		{
			gtk_box_pack_start(GTK_BOX(ihbox), pbtn, FALSE, FALSE, 0);
			gtk_box_pack_start(GTK_BOX(ihbox), dp->scwin, TRUE, TRUE, 0);
		}
		else
		{
			gtk_box_pack_start(GTK_BOX(ihbox), dp->scwin, TRUE, TRUE, 0);
			gtk_box_pack_start(GTK_BOX(ihbox), pbtn, FALSE, FALSE, 0);
		}
	}
	else
		gtk_box_pack_start(GTK_BOX(ihbox), dp->scwin, TRUE, TRUE, 0);

	if(fmt->path_above)
	{
		gtk_box_pack_start(GTK_BOX(dp->vbox), dp->notebook, FALSE, FALSE, 0);
		gtk_box_pack_start(GTK_BOX(dp->vbox), ihbox, TRUE, TRUE, 0);
	}
	else
	{
		gtk_box_pack_start(GTK_BOX(dp->vbox), ihbox, TRUE, TRUE, 0);
		gtk_box_pack_start(GTK_BOX(dp->vbox), dp->notebook, FALSE, FALSE, 0);
	}
	gtk_widget_show_all(dp->vbox);
	dp->dbclk_row = -1;
	dp->focus_row = -1;
	dp->old_focus_row = -1;
	clear_stats(dp);

	return dp->vbox;
}
